
/*
 put all the standard headers in GenCurveFit.h, including those from wavemetrics
 */
#include "XOPStandardHeaders.h"
#include "multiopenfiles.h"
#include <Carbon/Carbon.h>
	
// Structure used during the filtering of files.
struct NavFileTypeInfo {
	char extensions[256];			// e.g., ".txt,.itx"
};

typedef struct NavFileTypeInfo NavFileTypeInfo;
typedef struct NavFileTypeInfo* NavFileTypeInfoPtr;
typedef struct NavFileTypeInfo** NavFileTypeInfoHandle;

struct NavCallbackData {		// Data that is passed by Navigation Services to our filter and event functions.
	int gotNavStartMessage;						// Used to workaround a bug in Navigation Service in Mac OS X.

	int isOpen;									// True if Open File dialog.
	int isSave;									// True if Save File dialog.

	NavFileTypeInfoHandle ftiH;				// Used during filtering of files. Can be NULL.
	char *initialLocation;
	NavMenuItemSpecArrayHandle menuItemSpecsH;	// Description of our custom menu items. Can be NULL.

	int index;									// Index into ftiH and menuItemSpecsH to the selected type.
};
typedef struct NavCallbackData NavCallbackData;
typedef struct NavCallbackData* NavCallbackDataPtr;

static void NavFileEventProc(const NavEventCallbackMessage callBackSelector, NavCBRecPtr callBackParms, NavCallBackUserData callBackUD){
	NavCallbackDataPtr callbackDataPtr;
	callbackDataPtr = (NavCallbackDataPtr)callBackUD;

	switch(callBackSelector) {
		case kNavCBStart:
			{	// Set the initial item in the Show popup menu.
				NavMenuItemSpecArrayHandle menuItemSpecsH;
				int index, numMenuItemSpecs;
				menuItemSpecsH = callbackDataPtr->menuItemSpecsH;
				index = callbackDataPtr->index;
				if (menuItemSpecsH != NULL) {
					numMenuItemSpecs = GetHandleSize((Handle)menuItemSpecsH) / sizeof(NavMenuItemSpec);
					if (index>=0 && index<numMenuItemSpecs) {		// Sanity check.
						NavMenuItemSpec menuItemSpec;
						menuItemSpec = (*menuItemSpecsH)[index];
						NavCustomControl(callBackParms->context, kNavCtlSelectCustomType, &menuItemSpec);
					}
				}
				callbackDataPtr->gotNavStartMessage = 1;
			}
			break;
			
		case kNavCBEvent:
			switch (callBackParms->eventData.eventDataParms.event->what) {
				case updateEvt:
//					if (!IsMacOSX())	// Mac OS X does the updating automatically, apparently using offscreen bitmaps.
//						DoUpdate();		// Call back to Igor to do update.
					break;
			}
			break;
			
		case kNavCBPopupMenuSelect:		// User made selection in Show popup menu.
			if (callbackDataPtr->gotNavStartMessage) {		// In Mac OS X, we received this message before the dialog appears. This is a Mac OS X bug.
				NavMenuItemSpecPtr msp;
				
				// Remember the index into the array of NavMenuItemSpecs.
				msp = (NavMenuItemSpecPtr)callBackParms->eventData.eventDataParms.param;
				callbackDataPtr->index = msp->menuType;		// index is an index into our ftiH and menuItemSpecsH handles.
			}
			break;
	}
}


/*	NavGetMenuItemSpecsHandle(NavFileTypeInfoHandle ftiH, int numMenuItems)
	
	Creates and returns a NavMenuItemSpecArrayHandle which is passed to
	Mac Navigation Services to create the custom menu items.
	
	Returns NULL if error.
*/
 NavMenuItemSpecArrayHandle
NavGetMenuItemSpecsHandle(NavFileTypeInfoHandle ftiH, int numMenuItems)
{
	NavMenuItemSpecArrayHandle menuItemSpecsH;
	long numBytes;
	int i;
	
	numBytes = numMenuItems*sizeof(NavMenuItemSpec);
	menuItemSpecsH = (NavMenuItemSpecArrayHandle)NewHandle(numBytes);
	if (menuItemSpecsH == NULL)
		return NULL;

	MemClear(*menuItemSpecsH, numBytes);
	for(i=0; i<numMenuItems; i+=1) {
		char menuItemText[256];
		
		(*menuItemSpecsH)[i].version = kNavMenuItemSpecVersion;
		(*menuItemSpecsH)[i].menuCreator = kNavGenericSignature;
		(*menuItemSpecsH)[i].menuType = i;		// This is an index into menuItemSpecsH and ftiH.
		strcpy(menuItemText, (*ftiH)[i].extensions);
		CopyCStringToPascal(menuItemText, (*menuItemSpecsH)[i].menuItemName);
	}	
	return menuItemSpecsH;
}

/*	NavParseFileFilterString(fileFilterStr, ftiHPtr)
	
	Parses the fileFilterStr into a form more easily used by the file filter function.	
	
	If the fileFilterStr can be parsed correctly, NavParseFileFilterString
	returns a newly created NavFileTypeInfoHandle via ftiHPtr
	and returns a function result of 0. You must dispose of the handle
	when you are finished with it.
	
	If there is an error in parsing, no handle is returned and the function
	result is non-zero.
*/
static int NavParseFileFilterString(const char* fileFilterStr, NavFileTypeInfoHandle* ftiHPtr){
	NavFileTypeInfoHandle ftiH;
	NavFileTypeInfo fti;
	const char* p;
	const char* p1;
	
	*ftiHPtr = NULL;
	
	ftiH = (NavFileTypeInfoHandle)NewHandle(0L);
	
	p = fileFilterStr;
	while(*p != 0) {
		MemClear(&fti, sizeof(fti));
		
		// Now get the extensions associated with this menu item.
				
		p1 = strchr(p, ';');
		if (p1 == NULL) {
			DisposeHandle((Handle)ftiH);
			return 5;						// Error - no semicolon to mark end of extensions.
		}
		
		if (p1 - p >= sizeof(fti.extensions)) {
			DisposeHandle((Handle)ftiH);
			return 6;						// Error - extensions string too long.
		}
		
		strncpy(fti.extensions, p, p1-p);
		p = p1 + 1;							// Skip past extensions to next menu item, if any.
		
		// Add the NavFileTypeInfo record to the handle.
		{
			long origNumBytes = GetHandleSize((Handle)ftiH);
			SetHandleSize((Handle)ftiH, origNumBytes + sizeof(NavFileTypeInfo));
			if (MemError()) {
				DisposeHandle((Handle)ftiH);
				return 7;
			}
			memcpy((char*)*ftiH + origNumBytes, &fti, sizeof(NavFileTypeInfo));
		}
	}
	
	*ftiHPtr = ftiH;
	return 0;
}

static int NavFileDoFilter(NavCallbackDataPtr callbackDataPtr, const char* theName){
	NavFileTypeInfoHandle ftiH;
	NavFileTypeInfoPtr ftiP;
	int index, numMenuItems;
	const char* ep;					// Pointer to extensions.
	const char* p;
	
	ftiH = callbackDataPtr->ftiH;
	if (ftiH == NULL)
		return 1;
	
	index = callbackDataPtr->index;				// Zero-based index of currently-selected menu item.
	numMenuItems = GetHandleSize((Handle)ftiH) / sizeof(NavFileTypeInfo);
	if (index<0 || index>=numMenuItems)
		return 1;								// Something is wrong.

	ftiP = &(*ftiH)[callbackDataPtr->index];	// DEREFERENCE
	ep = ftiP->extensions;						// DEREFERENCE. Points to something like ".txt,.dat,.csv"
	
	// See if this file name passes the extension test.
	
	{
		int theNameLen;
		char theExtension[64];					// Extension from this file's name.
		int theExtensionLen;
		
		p = strstr(ep, ".*");	//if the extension list has .* we accept everything
		if(p)
			return 1;

		p = strrchr(theName, '.');				// Find last dot.
		if (p == NULL)
			return 0;							// File has no extension so there can be no match.
		
		theNameLen = strlen(theName);
		
		theExtensionLen = theName + theNameLen - p;
		if (theExtensionLen >= sizeof(theExtension))
			return 0;							// File has an outlandish extension.
		
		strncpy(theExtension, p, theExtensionLen);
		theExtension[theExtensionLen] = 0;
	
		p = ep;
					
		while(*p) {									// Look through list of file extensions for one that matches this file's extension.
			char thisExtension[64];
			int thisExtensionLen;
			const char* p1;
			
			p1 = strchr(p, ',');					// Point to separator between extensions.
			if (p1 == NULL)
				p1 = p + strlen(p);					// Point to null after last extension.
			
			thisExtensionLen = p1 - p;
			if (thisExtensionLen >= sizeof(thisExtension))
				return 0;							// Outlandish filter string.
			
			if (thisExtensionLen == theExtensionLen) {
				strncpy(thisExtension, p, thisExtensionLen);
				thisExtension[thisExtensionLen] = 0;
			
				if (CmpStr(thisExtension, theExtension) == 0)
					return 1;						// Extensions match.
			}
			
			if (*p1 == 0)
				break;								// End of extensions.
				
			p = p1 + 1;								// Skip comma and continue.
		}
	}
	
	return 0;			// Failed all tests.
}

static Boolean NavFileFilter(AEDesc* theItem, void* info, NavCallBackUserData callBackUD, NavFilterModes filterMode){
	int display;
	NavFileOrFolderInfo* theInfo;
	NavCallbackDataPtr callbackDataPtr;
		
	callbackDataPtr = (NavCallbackDataPtr)callBackUD;

	display = 1;
	
	theInfo = (NavFileOrFolderInfo*)info;

	if (theInfo->isFolder == true) // 1
		return display;
	
	AECoerceDesc (theItem, typeFSRef, theItem); // 2
	  
	if (theItem->descriptorType == typeFSRef) {
		if (!theInfo->isFolder) {
			FSRef ref;
			FSSpec spec;
			char theName[256];
			int err;

			*theName = 0;
			err = AEGetDescData(theItem, &ref, sizeof(FSRef));
			
			err = FSGetCatalogInfo(&ref, kFSCatInfoNone, NULL, NULL, &spec, NULL);

			if (err == 0)
				CopyPascalStringToC(spec.name, theName);
			
			display = NavFileDoFilter(callbackDataPtr, theName);
		}
	}
	
	return display;
}


int
Executemultiopenfiles(multiopenfilesRuntimeParamsPtr p){
	int err = 0;
	NavDialogCreationOptions dialogOptions;
	NavDialogRef dialog;
	NavReplyRecord replyRecord;
	NavUserAction userAction;

	char initialLocation[MAX_PATH_LEN+1];	
	AEDesc initLocation;
	AEDesc *initLocationPtr = NULL;
	FSSpec initialDirSpec;
	FSSpec* initialDirSpecPtr = NULL;
	
	NavCallbackData callbackData;
	NavFileTypeInfoHandle ftiH;

	int numMenuItems;
	char fileFilterStr[256];
	int *fileIndexPtr = NULL;					// NULL or one-based index into type list
	NavEventUPP eventUPP = NewNavEventUPP(NavFileEventProc);
	NavObjectFilterUPP filterUPP = NewNavObjectFilterUPP(NavFileFilter);
	NavMenuItemSpecArrayHandle menuItemSpecsH;
	CFMutableArrayRef fileTypes = NULL;

	CFURLRef fileAsCFURLRef = NULL;
	FSRef fileAsFSRef;
	OSStatus status;
	long theCount, ii;
	long V_flagRes = -1;
	CFStringRef windowTitle = NULL;
	char windowStr[256];
	
	Boolean succeeded;
	char urlOut[MAX_PATH_LEN+1];
	Handle urlOutHand = NULL;
	int cancelled = 0;
		
	memset(initialLocation,0,sizeof(initialLocation));
	// Flag parameters.
	urlOutHand = NewHandle(0);
	if(urlOutHand == NULL){
		err = NOMEM;
		goto done;
	}
	
	if (p->FFlagEncountered) {
		// Parameter: p->FFlagStrH (test for NULL handle before using)
		if(p->FFlagStrH==NULL){
			err = OH_EXPECTED_STRING;
			goto done;
		}
		if(err = GetCStringFromHandle(p->FFlagStrH, fileFilterStr, 255))
			goto done;
	}

	if (p->PFlagEncountered) {
		// Parameter: p->PFlagName
		if(err = GetPathInfo2(p->PFlagName, initialLocation))
			goto done;
	}

	// Prepare structure that is passed by Navigation Services to our filter and event functions.
	MemClear(&callbackData, sizeof(callbackData));
	callbackData.isOpen = 1;
	callbackData.initialLocation = initialLocation;
	
	// Parse filter strings, if any. This defines Show menu items.
	ftiH = NULL;
	numMenuItems = 0;
	if (fileFilterStr != 0) {								// Not empty fileFilterStr?
		NavParseFileFilterString(fileFilterStr, &ftiH);	// ftiH will be NULL if fileFilterStr is not correctly formed.
		if (ftiH != NULL)
			numMenuItems = GetHandleSize((Handle)ftiH) / sizeof(NavFileTypeInfo);
	}
	callbackData.ftiH = ftiH;
	
	// Get the standard set of defaults
	status = NavGetDefaultDialogCreationOptions(&dialogOptions);
	require_noerr( status, CantGetNavOptions );
	
	// If we are adding custom items to the Show menu, create structure to pass to Navigation services.
	menuItemSpecsH = NULL;
	if (numMenuItems>0 && ftiH!=NULL) {				// We want do filter files?
		menuItemSpecsH = NavGetMenuItemSpecsHandle(ftiH, numMenuItems);
		int jj;
		fileTypes = CFArrayCreateMutable(NULL, numMenuItems, &kCFTypeArrayCallBacks);

		for(jj=0 ; jj<numMenuItems ; jj+=1){
			CFStringRef temp = CFStringCreateWithCString(NULL, (*ftiH)[jj].extensions, kCFStringEncodingUTF8);
			CFArrayAppendValue(fileTypes, temp );
			CFRelease(temp);
		}
		dialogOptions.popupExtension = fileTypes;

		/*	This is what will be initially selected in the Show popup menu.
			By setting these fields, we make our filter function aware of that fact.
		*/
		if (fileIndexPtr != NULL) {						// If fileIndexPtr not null, honor the caller's request.
			if (*fileIndexPtr > 0) {					// *fileIndexPtr is one-based.
				if (*fileIndexPtr <= numMenuItems)
					callbackData.index = *fileIndexPtr-1;
			}
		}
	}
	callbackData.menuItemSpecsH = menuItemSpecsH;			// Can be NULL. If so, Navigation Services will display all files.
		
	if (p->MFlagEncountered) {
		// Parameter: p->MFlagStrH (test for NULL handle before using)
		if(p->MFlagStrH == NULL){
			err = OH_EXPECTED_STRING;
			goto done;
		}
		if(err = GetCStringFromHandle(p->MFlagStrH, windowStr, 255))
			goto done;
		windowTitle = CFStringCreateWithBytes (NULL, (UInt8*)windowStr, strlen(windowStr),  CFStringGetSystemEncoding(), false);
		dialogOptions.windowTitle = windowTitle;
	}
	
	// Make the window app-wide modal
	dialogOptions.modality = kWindowModalityAppModal;
	
	// Create the dialog
	status = NavCreateGetFileDialog(&dialogOptions, NULL, eventUPP, NULL, filterUPP, &callbackData, &dialog);
	require_noerr( status, CantCreateDialog );

	if(strlen(initialLocation)){
		initLocationPtr = NULL;
		unsigned char temp[MAX_PATH_LEN];
		CopyCStringToPascal(initialLocation, temp);
		if (err = FSMakeFSSpec(0, 0, temp, &initialDirSpec)) {
				IgorError("XOPOpenFileDialog FSMakeFSSpec", err);
				goto done;
			}
			initialDirSpecPtr = &initialDirSpec;
		
		if (initialDirSpecPtr != NULL) {
			err = AECreateDesc(typeFSS, initialDirSpecPtr, sizeof(FSSpec), &initLocation);
		if (err == 0)
			initLocationPtr = &initLocation;
		}
	
		status = NavCustomControl(dialog, kNavCtlSetLocation, initLocationPtr);
		require((status == noErr), done);
	}
	
	// Show it
	status = NavDialogRun(dialog);
	require_noerr( status, CantRunDialog );
	
	userAction = NavDialogGetUserAction(dialog);
	if(userAction == kNavUserActionCancel || userAction == kNavUserActionNone)
		goto CantGetReply;
	
	// Get the reply
	status = NavDialogGetReply(dialog, &replyRecord);
	require( ((status == noErr) || (status == userCanceledErr)), CantGetReply );
	
	// If the user clicked "Cancel", just bail
	if ( status == userCanceledErr ){
		cancelled = 1;
		goto UserCanceled;
	}
	
	status = AECountItems(&(replyRecord.selection), &theCount);
	require((status == noErr), done);
	
	for(ii=1 ; ii<theCount+1 ; ii+=1){		
		if(ii>1){
			if(err = PtrAndHand(";", urlOutHand, 1))
				goto done;
		}
		// Get the file
		status = AEGetNthPtr(&(replyRecord.selection), ii, typeFSRef, NULL, NULL, &fileAsFSRef, sizeof(FSRef), NULL);
		require_noerr( status, CantExtractFSRef );
		
		// Convert it to a CFURL, then to a CFString string representation
		fileAsCFURLRef = CFURLCreateFromFSRef(NULL, &fileAsFSRef);
		CFStringRef url = CFURLCopyFileSystemPath (fileAsCFURLRef, kCFURLHFSPathStyle);
	
		CFRelease(fileAsCFURLRef); // release when finished with it.
		fileAsCFURLRef = NULL;
		
		succeeded = CFStringGetCString(url, urlOut, MAX_PATH_LEN, CFStringGetSystemEncoding()); 
		CFRelease(url);
		 		
		V_flagRes = 0;
		if(err = PtrAndHand(urlOut, urlOutHand, strlen(urlOut)))
			goto done;
	}
	
	//////////CLEANUP///////////////	
done:

CantExtractFSRef:
	
UserCanceled:
	verify_noerr( NavDisposeReply(&replyRecord) );
CantGetReply:
	
CantRunDialog:
	NavDialogDispose(dialog);
CantCreateDialog:
	
CantGetNavOptions:
	if(initLocationPtr)
		AEDisposeDesc(initLocationPtr);	
	if(fileAsCFURLRef)
		CFRelease(fileAsCFURLRef); // release when finished with it.
		
	if(windowTitle)
		CFRelease(windowTitle);
	if (eventUPP != NULL)
		DisposeNavEventUPP(eventUPP);
	if (filterUPP != NULL)
		DisposeNavObjectFilterUPP(filterUPP);
	if (menuItemSpecsH != NULL)
		DisposeHandle((Handle)menuItemSpecsH);
	if (ftiH != NULL)
		DisposeHandle((Handle)ftiH);
	if (fileTypes)
		CFRelease(fileTypes);
	
	if(err || cancelled){
		V_flagRes=-1;
		SetOperationStrVar("S_filename", "");
	}

	err = SetOperationNumVar("V_flag", V_flagRes);
		
	if (GetHandleSize(urlOutHand) > 0 && !V_flagRes)
		err = SetOperationStrVar2("S_filename", *urlOutHand, GetHandleSize(urlOutHand));
	
	if(urlOutHand)
		DisposeHandle(urlOutHand);
	
	return err;
}